"use strict";
const entities = require('entities');
const xpath = require('xpath');
const path = require('path');
const moment = require('moment-timezone');
const dom = require('xmldom').DOMParser;
const EventEmitter = require('events');


/**
 * General Parser functions
 * @type {Parser}
 */
module.exports = class Parser extends EventEmitter{
    
    constructor(config, key) {
                
        super();
        
        this.config = config;
        this.key = key;
        
        
    }

    /**
     * 
     * Do general processing on an item
     * 
     * @param itemData
     * @param currentModelType
     * @returns {*}
     */
    processItem(itemData, currentModelType) {


        //Check the required fields
        let required = ["duration", "external_id", "model_type_id"];
        required.forEach(function(req) {
            if (typeof itemData[req] == "undefined" ){
                throw Error('Missing fields, Required fields are ' + required.toString());
            }
        });
        
        if (itemData["model_type_id"] === 0 || itemData["model_type_id"] === '0') {
            //model_type_id is defined as 0, we can ignore this itemtype
            return false;
        }
        
        //Extra data stuffz, first duration to seconds
        if (this.config.duration_ms) {
            itemData["duration"] = Math.round(parseInt(itemData["duration"]) / 1000);
        }
        else {
            itemData["duration"] = this.durationToSeconds(itemData["duration"]);
        }

        //Check and possibly fix the item start time
        if (itemData["start"]) {
            //Parse the start date of the item, possibly with forced format/timexzone
            let parsedStart = this.parseDateTime(itemData["start"]);
            
            //adapt the start date of the item, if global date is defined
            if (this.globalDate) {
                if (parsedStart) {
                    itemData["start"] = this.parseDateTime(this.globalDate.format("YYYY-MM-DD") + " " + parsedStart.format("HH:mm:ss"), true);
                } else {
                    itemData["start"] = this.parseDateTime(this.globalDate.format("YYYY-MM-DD")+ " " +itemData["start"], true);
                }
            } else if (parsedStart) {
                itemData["start"] = parsedStart;
            } else {
                //invalid date for start, cannot fix it with global date
                delete itemData["start"];
                if (this.config.type == 'now-on-air') {
                    throw Error('Can not parse "start" which is required for now playing')
                }
            }

        } else if (this.config.type == 'now-on-air') {
            //this is no playlist, but there is no start time
            if (this.globalDate) {
                itemData["start"] = this.globalDate;
            } else {
                throw Error('Missing field "start" which is required for now playing')
            }
        }

        itemData = this.addItemDataFieldEnrichtments(itemData, currentModelType);

        itemData = this.duplicateCustomDataToFieldValues(itemData);

        return itemData; 
    }


    /**
     * 
     * Put non-standard fields inside the field_values
     * 
     * @param itemData
     * @returns {*}
     */
    duplicateCustomDataToFieldValues(itemData) {
        let dontMove = ['title', 'start', 'duration', 'external_id', 'field_values', 'model_type_id', '_start_seconds', '_rawhour_block'];
        var field_values = {};
        for (let key in itemData){
            if (dontMove.indexOf(key) === -1){
                field_values[key] = itemData[key];
            }
        }
        if (Object.keys(field_values).length) {
            itemData.field_values = field_values;
        }
        return itemData;
    }

    /**
     * 
     * Apply the field enrichments (combine or add data to the fields)
     * 
     * @param itemData
     * @param currentModelType
     * @returns {*}
     */
    
    addItemDataFieldEnrichtments(itemData, currentModelType) {
        //add the data enrichments
        if (typeof currentModelType.data_field_enrichments != "undefined") {
            var enrichments = currentModelType.data_field_enrichments;
            for (let key in enrichments) {
                var newValue = enrichments[key];
                for (let field in itemData) {
                    newValue = newValue.replace('{'+field+'}', itemData[field])
                }
                itemData[key] = newValue;
            }
        }
        return itemData
    }


    /**
     * 
     * Do the modeltype matching
     * 
     * @param value
     * @param modelType
     * @returns {boolean}
     */
    matchModelType(value, modelType) {
        let match = modelType.model_type_match;
        let type = modelType.model_type_match_type;

        if (type == "full" && value == match) {
            return true;
        }

        if (type == "contains" && value.indexOf(match) != -1) {
            return true;
        }
        if (type == "regex") {
            var rePattern = new RegExp(match);
            if (value.match(rePattern)) {
                return true;
            }
        }
        
        return;
        
    }

    /**
     * durationToSeconds
     *
     * @param {String} duration
     * @returns {Integer}
     */
    durationToSeconds(duration) {
        if (typeof duration == "number") {
            return duration;
        }
        
        let p = duration.split(':'),
            s = 0,
            m = 1;

        while (p.length > 0) {
            s += m * parseInt(p.pop(), 10);
            m *= 60;
        }

        return s;
    }

    
    /**
     * 
     * get the global date out of the filename, or an xpath (XML specific)
     * 
     * @param doc
     * @returns {*}
     */
    getGlobalDate() {

        let date = false;
        if (typeof this.config.global_date_filename != "undefined" && this.config.global_date_filename) {
            let format = false;
            if (typeof this.config.global_date_filename == "string") {
                format = this.config.global_date_filename;
            }
            date = this.parseDateTime(this.filename, format);
        } else if (typeof this.config.global_date_xpath != "undefined" && this.config.global_date_xpath ) {
            const dateNode = xpath.select1(this.config.global_date_xpath, this.doc);
            if (!dateNode) throw Error('Invalid Global Date with selector: ' + this.config.global_date_xpath);
            date = this.parseDateTime(this.getNodeValue(dateNode));
        }
        if (typeof date != "object") {
            return false;
        }
        
        this.globalDate = date;
        
        return date;
    }
    
    
    /**
     *
     * Parse a string to a moment object, with format/timezone depending on configuration
     * Fallback without format/timezone if moment is invalid
     * 
     * @param string
     * @param useFormat
     * @returns {boolean}
     *
     */
    parseDateTime(string, overrideFormat) {
        var format = "";
        if (typeof overrideFormat == "undefined") {
            //no override format
            overrideFormat = false;
        } else if (typeof overrideFormat == "string") {
            //explicit override format, use this instead
            format = overrideFormat;
        } else if (overrideFormat === true) {
            //overide format is true, dont use the config format
            format = false;
        }
        
        //if overide format is false or undefined, try to get the format from config
        if (!overrideFormat){
            format = (this.config.date_format && this.config.date_format != "") ? this.config.date_format : false;
        }
        var timezone = (this.config.timezone && this.config.timezone != "") ? this.config.timezone : false;
        var output = false;

        try {
            if (format && timezone) {
                //output with predefined format AND timezone
                output = moment.tz(string, format, timezone);
            }
            if ((!format && timezone) || (output && !output.isValid())) {
                //output with predefined timezone ONLY, fallback for format AND timezone attempt
                output = moment.tz(string, timezone);
            }
            if ((!timezone && format) || (format && output && !output.isValid())) {
                //output with predefined format ONLY  fallback for format AND timezone attempts
                output = moment(string, format);
            }
            if (!output || !output.isValid()) {
                //output without predefined format or timezone, fallback for all
                output = moment(string);
            }
        } catch (e) {
            return false;
        }
        
        return output.isValid() ? output : false;

    }

    /**
     * Emit a non fatal warging for the logs
     * 
     * @param msg
     */
    
    emitWarning(msg) {
        console.log("[WARN] Parser Warning: "+msg)
        this.emit('error', "[WARN] Parser Warning: "+msg);
    }
    
}